"""
An archive, for instance a zip or tar file, that contains
an arbritrary collection of files (e.g. model weights).
"""

from typing import Literal

from hippometa.base import BaseMetadata


class ArchiveMetadata(BaseMetadata):
    """
    Metadata for an archive object.
    """

    metadata_type: Literal["archive"] = "archive"

    # The type of archive, e.g. zip, tar, etc.
    archive_type: str



"""
Core client object for interacting with the hippo API.
"""

import httpx
from pydantic_settings import (
    BaseSettings,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from rich.console import Console
from soauth.toolkit.client import SOAuth

from .caching import Cache, MultiCache

console = Console()


def Client(
    host: str,
    token_tag: str | None,
    timeout: int | None = None,
) -> httpx.Client:
    auth = SOAuth(token_tag) if token_tag else None
    return httpx.Client(base_url=host, auth=auth, timeout=httpx.Timeout(timeout or 60))


class ClientSettings(BaseSettings):
    """
    Main settings for the hippo API client. Used to configure:

    1. API access (key and host)
    2. Caching (path to cache(s))
    3. Verbosity.
    """

    host: str = "http://localhost:8000"
    "The hostname of the HIPPO service you are connected to"
    token_tag: str | None = None
    "The tag associated with the API key for the hippo API to use"
    verbose: bool = False
    "Verbosity control: set to true for extra info"
    use_multipart_upload: bool = False
    "The size of the multipart upload parts in bytes. If set to zero, no multipart uploads will be used"
    client_timeout: int = 60
    "The timeout for the client in seconds. Default is 60 seconds (up from the httpx default of 5)"

    default_readers: list[str] = []
    "Default readers for new collections and products"
    default_writers: list[str] = []
    "Default writers for new collections and products"

    caches: list[Cache] = []

    model_config = SettingsConfigDict(
        json_file=("config.json", "~/.hippo.conf"), env_prefix="hippo_"
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: BaseSettings,
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            dotenv_settings,
            env_settings,
            JsonConfigSettingsSource(settings_cls),
            file_secret_settings,
        )

    @property
    def cache(self) -> MultiCache:
        """
        Return a MultiCache object for the caches.
        """
        return MultiCache(caches=self.caches)

    @property
    def client(self) -> Client:
        """
        Return a Client object for the API.
        """
        return Client(
            token_tag=self.token_tag, host=self.host, timeout=self.client_timeout
        )



"""
The storage access layer for hippo. Assumes that you have an S3-compatible backend,
and uses the MinIO tools.
"""

import datetime
import os
from math import ceil

from loguru import logger
from minio import Minio
from minio.api import Part, genheaders
from minio.error import S3Error
from pydantic import BaseModel, ConfigDict


def replace_host(url: str, old: str, new: str | None, upgrade: bool) -> str:
    """
    Replaces the host in a URL.
    """

    if new is not None:
        url = url.replace(old, new)

    if upgrade:
        url = url.replace("http://", "https://")

    return url


class Storage(BaseModel):
    url: str
    access_key: str
    secret_key: str
    presign_url: str | None = None

    upgrade_presign_url_to_https: bool = False
    secure: bool = False
    cert_check: bool = False

    client: Minio | None = None
    expires: datetime.timedelta = datetime.timedelta(days=1)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, __context):
        self.client = Minio(
            self.url,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=self.secure,
            cert_check=self.cert_check,
        )

    def object_name(self, filename: str, uploader: str, uuid: str) -> str:
        # Filename may contain some kind of path, and this
        # breaks the download side of things, but for some
        # reason this is a valid object name for uploads?
        return f"{uploader}/{uuid}/{os.path.basename(filename)}"

    def bucket(self, name: str):
        if not self.client.bucket_exists(name):
            self.client.make_bucket(name)

        return

    def put(
        self, name: str, uploader: str, uuid: str, bucket: str, size: int, batch: int
    ) -> tuple[str, list[str]]:
        """
        Initiates a multipart upload. Returns a list of pre-signed URLs and the
        upload ID. Each one of these URLs should have ``batch`` data uploaded to
        it.

        Returns
        -------
        upload_id : str
            The upload ID for the multipart upload.
        urls : list[str]
            A list of pre-signed URLs to upload data to.
        """

        self.bucket(name=bucket)

        object_name = self.object_name(
            filename=name,
            uploader=uploader,
            uuid=uuid,
        )

        # First - initiate the multipart upload to get the UploadID.
        upload_id = self.client._create_multipart_upload(
            bucket_name=bucket,
            object_name=object_name,
            headers=genheaders(None, None, None, None, None),
        )

        # Second - generate n pre-signed URLs for each part.
        urls = [
            replace_host(
                url=self.client.get_presigned_url(
                    method="PUT",
                    bucket_name=bucket,
                    object_name=object_name,
                    expires=self.expires,
                    # Parts must use one-based indexing (s3 requirement)
                    extra_query_params={
                        "partNumber": str(i + 1),
                        "uploadId": upload_id,
                    },
                ),
                old=self.url,
                new=self.presign_url,
                upgrade=self.upgrade_presign_url_to_https,
            )
            for i in range(int(ceil(size / batch)))
        ]

        return upload_id, urls

    def complete(
        self,
        name: str,
        uploader: str,
        uuid: str,
        bucket: str,
        upload_id: str,
        headers: list[dict[str, str]],
        sizes: list[int],
    ) -> None:
        """
        Completion after a multipart upload. If this doesn't occur, a multipart upload will not
        succeed. Requires information from the client (list[Part]) that uploaded the data.
        """

        self.bucket(bucket)

        self.client._complete_multipart_upload(
            bucket_name=bucket,
            object_name=self.object_name(
                filename=name,
                uploader=uploader,
                uuid=uuid,
            ),
            upload_id=upload_id,
            parts=[
                # Sometimes the s3 server responds with ETag, sometimes etag. Lol.
                Part(
                    part_number=i + 1,
                    etag={x.lower(): y for x, y in headers[i].items()}["etag"],
                    size=sizes[i],
                )
                for i in range(len(headers))
            ],
        )

    def confirm(self, name: str, uploader: str, uuid: str, bucket: str) -> bool:
        """
        Checks whether an object exists.
        """

        self.bucket(name=bucket)

        try:
            self.client.get_object_tags(
                bucket_name=bucket,
                object_name=self.object_name(
                    filename=name,
                    uploader=uploader,
                    uuid=uuid,
                ),
            )

            logger.debug(
                f"Object {self.object_name(filename=name, uploader=uploader, uuid=uuid)} exists."
            )

            return True
        except S3Error as e:
            logger.info(e)
            return False

    def get(self, name: str, uploader: str, uuid: str, bucket: str) -> str:
        """
        Returns a specific URL for HTTP GET requests.
        """

        self.bucket(name=bucket)

        # Special headers - if we have a video or image, we want it to
        # be displayed inline in the browser, not downloaded.

        match x := name.split(".")[-1].lower():
            case "jpg" | "jpeg" | "png" | "gif":
                response_headers = {
                    "response-content-type": f"image/{x}",
                    "response-content-disposition": "inline",
                }
            case "mp4" | "webm":
                response_headers = {
                    "response-content-type": f"video/{x}",
                    "response-content-disposition": "inline",
                }
            case _:
                response_headers = None

        base_url = self.client.presigned_get_object(
            bucket_name=bucket,
            object_name=self.object_name(
                filename=name,
                uploader=uploader,
                uuid=uuid,
            ),
            expires=self.expires,
            response_headers=response_headers,
        )

        return replace_host(
            base_url,
            old=self.url,
            new=self.presign_url,
            upgrade=self.upgrade_presign_url_to_https,
        )

    def delete(self, name: str, uploader: str, uuid: str, bucket: str) -> str:
        """
        Deletes an object from the bucket.
        """

        self.client.remove_object(
            bucket_name=bucket,
            object_name=self.object_name(
                filename=name,
                uploader=uploader,
                uuid=uuid,
            ),
        )



"""
API endpoints for relationships between products and collections.
"""

from beanie import PydanticObjectId
from fastapi import APIRouter, HTTPException, Request, status
from loguru import logger

from hipposerve.api.models.relationships import (
    CreateCollectionRequest,
    ReadCollectionCollectionResponse,
    ReadCollectionProductResponse,
    ReadCollectionResponse,
    UpdateCollectionRequest,
)
from hipposerve.service import acl, collection, product
from hipposerve.service.auth import requires

relationship_router = APIRouter(prefix="/relationships")


@relationship_router.put("/collection/{name}")
@requires(["hippo:admin", "hippo:write"])
async def create_collection(
    name: str,
    model: CreateCollectionRequest,
    request: Request,
) -> PydanticObjectId:
    """
    Create a new collection with {name}.
    """

    logger.info(
        "Request to create collection: {} from {}", name, request.user.display_name
    )

    coll = await collection.create(
        name=name,
        user=request.user.display_name,
        description=model.description,
        collection_readers=model.readers,
        collection_writers=model.writers,
    )

    logger.info(
        "Collection {} ({}) created for {}", coll.id, name, request.user.display_name
    )

    return coll.id


@relationship_router.get("/collection/{id}")
@requires(["hippo:admin", "hippo:read"])
async def read_collection(
    id: PydanticObjectId,
    request: Request,
) -> ReadCollectionResponse:
    """
    Read a collection's details.
    """

    logger.info("Request to read collection: {} from {}", id, request.user.display_name)

    try:
        item = await collection.read(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    return ReadCollectionResponse(
        id=item.id,
        name=item.name,
        description=item.description,
        owner=item.owner,
        readers=item.readers,
        writers=item.writers,
        products=[
            ReadCollectionProductResponse(
                id=x.id,
                name=x.name,
                version=x.version,
                description=x.description,
                owner=x.owner,
                uploaded=x.uploaded,
                metadata=x.metadata,
            )
            for x in item.products
        ],
        child_collections=[
            ReadCollectionCollectionResponse(
                id=x.id,
                name=x.name,
                description=x.description,
                owner=x.owner,
                readers=x.readers,
                writers=x.writers,
            )
            for x in item.child_collections
        ],
        parent_collections=[
            ReadCollectionCollectionResponse(
                id=x.id,
                name=x.name,
                description=x.description,
                owner=x.owner,
                readers=x.readers,
                writers=x.writers,
            )
            for x in item.parent_collections
        ],
    )


@relationship_router.get("/collection/search/{name}")
@requires(["hippo:admin", "hippo:read"])
async def search_collection(
    name: str,
    request: Request,
) -> list[ReadCollectionResponse]:
    """
    Search for collections by name. Products and sub-collections are not
    returned; these should be fetched separately through the read_collection
    endpoint.
    """

    logger.info(
        "Request to search for collection: {} from {}", name, request.user.display_name
    )

    results = await collection.search_by_name(
        name=name, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Found {} collections for {} from {}",
        len(results),
        name,
        request.user.display_name,
    )

    return [
        ReadCollectionResponse(
            id=item.id,
            name=item.name,
            description=item.description,
            owner=item.owner,
            readers=item.readers,
            writers=item.writers,
            products=None,
        )
        for item in results
    ]


@relationship_router.put("/collection/{collection_id}/{product_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_product_to_collection(
    collection_id: PydanticObjectId,
    product_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a product to a collection.
    """

    logger.info(
        "Request to add product {} to collection {} from {}",
        product_id,
        collection_id,
        request.user.display_name,
    )

    try:
        coll = await collection.read(
            id=collection_id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    try:
        item = await product.read_by_id(
            id=product_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.add_collection(
            product=item,
            access_groups=request.user.groups,
            collection=coll,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully added {} to collection {}", item.name, coll.name)
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.delete("/collection/{collection_id}/{product_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_product_from_collection(
    collection_id: PydanticObjectId,
    product_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a product from a collection.
    """

    logger.info(
        "Request to remove product {} from collection {} from {}",
        product_id,
        collection_id,
        request.user.display_name,
    )

    try:
        coll = await collection.read(
            id=collection_id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )

    try:
        item = await product.read_by_id(
            id=product_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.remove_collection(
            product=item,
            access_groups=request.user.groups,
            collection=coll,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully removed {} from collection {}", item.name, coll.name)
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.post("/collection/{id}/diff")
@requires(["hippo:admin", "hippo:write"])
async def diff_collection(
    id: PydanticObjectId,
    model: UpdateCollectionRequest,
    request: Request,
) -> dict[str, list[str]]:
    """
    Diff a collection's current state with the proposed changes.
    """

    logger.info("Request to diff collection: {} from {}", id, request.user.display_name)

    diff = await collection.diff(
        id=id,
        access_groups=request.user.groups,
        scopes=request.auth.scopes,
        name=model.name,
        description=model.description,
    )

    logger.info(
        "Successfully diffed collection {} from {}", id, request.user.display_name
    )
    return diff


@relationship_router.post("/collection/{id}")
@requires(["hippo:admin", "hippo:write"])
async def update_collection(
    id: PydanticObjectId,
    model: UpdateCollectionRequest,
    request: Request,
) -> PydanticObjectId:
    logger.info(
        "Request to update collection: {} from {}", id, request.user.display_name
    )
    coll = await collection.update(
        id=id,
        access_groups=request.user.groups,
        scopes=request.auth.scopes,
        name=model.name,
        description=model.description,
    )

    coll = await acl.update_access_control(
        doc=coll,
        owner=model.owner,
        add_readers=model.add_readers,
        remove_readers=model.remove_readers,
        add_writers=model.add_writers,
        remove_writers=model.remove_writers,
    )

    logger.info(
        "Successfully updated collection {} from {}", id, request.user.display_name
    )
    return coll.id


@relationship_router.delete("/collection/{id}")
@requires(["hippo:admin", "hippo:write"])
async def delete_collection(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Delete a collection.
    """

    logger.info(
        "Request to delete collection: {} from {}", id, request.user.display_name
    )

    try:
        # Check if we have a parent; if we do, we need to remove its link to us.
        coll = await collection.read(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )

        if coll.parent_collections:
            for parent in coll.parent_collections:
                await collection.remove_child(
                    parent_id=parent.id,
                    child_id=id,
                    groups=request.user.groups,
                    scopes=request.auth.scopes,
                )

        await collection.delete(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        logger.info(
            "Successfully deleted collection {} from {}", id, request.user.display_name
        )
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.put("/product/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_child_product(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a child product to a parent product.
    """

    logger.info(
        "Request to add product {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        source = await product.read_by_id(
            id=parent_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        destination = await product.read_by_id(
            id=child_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.add_relationship(
            source=source,
            destination=destination,
            access_groups=request.user.groups,
            type="child",
            scopes=request.auth.scopes,
        )
        logger.info(
            "Successfully added {} as child of {}", destination.name, source.name
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.delete("/product/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_child_product(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a parent-child relationship between two products.
    """

    logger.info(
        "Request to remove product {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        source = await product.read_by_id(
            id=parent_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        destination = await product.read_by_id(
            id=child_id, groups=request.user.groups, scopes=request.auth.scopes
        )
        await product.remove_relationship(
            source=source,
            destination=destination,
            access_groups=request.user.groups,
            type="child",
            scopes=request.auth.scopes,
        )
        logger.info(
            "Successfully removed {} as child of {}", destination.name, source.name
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )


@relationship_router.put("/collection/{parent_id}/child_of/{child_id}")
@requires(["hippo:admin", "hippo:write"])
async def add_child_collection(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Add a child collection to a parent collection.
    """

    logger.info(
        "Request to add collection {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        await collection.add_child(
            parent_id=parent_id,
            child_id=child_id,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully added {} as child of {}", child_id, parent_id)
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.delete("/collection/{child_id}/child_of/{parent_id}")
@requires(["hippo:admin", "hippo:write"])
async def remove_child_collection(
    parent_id: PydanticObjectId,
    child_id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Remove a parent-child relationship between two collections.
    """

    logger.info(
        "Request to remove collection {} as child of {} from {}",
        child_id,
        parent_id,
        request.user.display_name,
    )

    try:
        await collection.remove_child(
            parent_id=parent_id,
            child_id=child_id,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        logger.info("Successfully removed {} as child of {}", child_id, parent_id)
    except collection.CollectionNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found."
        )


@relationship_router.post("/collection/{id}/pin")
@requires(["hippo:admin"])
async def pin_collection(
    id: PydanticObjectId,
    request: Request,
) -> PydanticObjectId:
    """
    Pin a collection.
    """

    logger.info("Request to pin collection: {} from {}", id, request.user.display_name)

    coll = await collection.pin(
        id=id, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully pinned collection {} from {}", id, request.user.display_name
    )
    return coll.id


@relationship_router.post("/collection/{id}/unpin")
@requires(["hippo:admin"])
async def unpin_collection(
    id: PydanticObjectId,
    request: Request,
) -> PydanticObjectId:
    """
    Unpin a collection.
    """

    logger.info(
        "Request to unpin collection: {} from {}", id, request.user.display_name
    )

    coll = await collection.unpin(
        id=id, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully unpinned collection {} from {}", id, request.user.display_name
    )
    return coll.id



from itertools import chain
from pathlib import Path
from typing import Iterable

import httpx
from pydantic import BaseModel, Field
from rich.console import Console

from henry.product import ProductInstance, RemoteProduct
from hippoclient import collections, relationships
from hippoclient.caching import MultiCache
from hipposerve.api.models.relationships import ReadCollectionResponse

from .exceptions import (
    CollectionIncompleteError,
    PreflightFailedError,
    ProductIncompleteError,
)


class CollectionInstance(BaseModel):
    pass


class LocalCollection(CollectionInstance):
    """
    A local coollection, created before pushing up to HIPPO. Can include
    LocalProduct instances that also haven't been pushed up, and other
    LocalCollection instances.
    """

    collection_id: str | None = None
    name: str
    description: str
    products: list[ProductInstance] = []
    collections: list[CollectionInstance] = []
    readers: list[str] = []
    writers: list[str] = []

    def __get_global_index(self, key: int, /) -> tuple[str, int]:
        if key < 0:
            return self.__get_global_index(self.__len__() + key)
        elif key < len(self.products):
            return "p", key
        elif key < self.__len__():
            return "c", key - len(self.products)
        else:
            raise IndexError(f"Index {key} invalid for object of size {self.__len__()}")

    def __repr__(self):
        return (
            f"LocalCollection(name='{self.name}', description='{self.description}', "
            f"products=[{', '.join([x.__repr__() for x in self.products])}], "
            f"collections=[{', '.join([x.__repr__() for x in self.collections])}])"
        )

    def __str__(self):
        return (
            f"{self.name} ({self.description}) collection with {len(self.products)} "
            f"products and {len(self.collections)} child collections"
        )

    def __len__(self) -> int:
        return len(self.products) + len(self.collections)

    def __getitem__(self, key: int, /) -> CollectionInstance | ProductInstance:
        arr, index = self.__get_global_index(key)
        if arr == "p":
            return self.products[index]
        else:
            return self.collections[index]

    def __iter__(self) -> Iterable[CollectionInstance | ProductInstance]:
        return chain(self.products, self.collections)

    def __reversed__(self) -> Iterable[ProductInstance | CollectionInstance]:
        return chain(self.collections.__reversed__(), self.products.__reversed__())

    def __contains__(self, item: CollectionInstance | ProductInstance, /):
        match item:
            case CollectionInstance():
                self.collections.__contains__(item)
            case ProductInstance():
                self.products.__contains__(item)
            case _:
                raise TypeError(
                    f"Only collections or products are members of collections, found {type(item)}"
                )

    def append(self, value: CollectionInstance | ProductInstance, /):
        match value:
            case CollectionInstance():
                self.collections.append(value)
            case ProductInstance():
                self.products.append(value)
            case _:
                raise TypeError(
                    f"Only collections or products may be appended to a collection, found {type(value)}"
                )

    def extend(self, iterable: Iterable[CollectionInstance | ProductInstance], /):
        for a in iterable:
            self.append(a)

    def pop(self, key: int, /) -> CollectionInstance | ProductInstance:
        arr, index = self.__get_global_index(key)
        if arr == "p":
            return self.products.pop(index)
        else:
            return self.collections.pop(index)

    def remove(self, value, /):
        match value:
            case CollectionInstance():
                self.collections.remove(value)
            case ProductInstance():
                self.products.remove(value)
            case _:
                raise TypeError(
                    f"Only collections or products may be part of a collection, found {type(value)}"
                )

    def preflight(self):
        """
        Run a pre-flight check on this collection. This checks that:

        a) The collection has a valid name and description.
        b) Runs preflight checks for all connected products.
        c) Runs preflight checks for all connected collections.
        """

        if self.collection_id:
            # We've already flown!
            return

        for preflightable in self:
            preflightable.preflight()

        if len(self.name) < 2 or not isinstance(self.name, str):
            raise PreflightFailedError(
                f"Name: {self.name} is not a valid name; ensure it is at least 2 characters and a valid string"
            )

        if len(self.description) < 2 or not isinstance(self.description, str):
            raise PreflightFailedError(
                f"Description: {self.description} is not a valid description; ensure it is at least 2 characters and a valid string"
            )

    def _upload(
        self,
        client: httpx.Client,
        console: Console,
        skip_preflight: bool = False,
        readers: list[str] | None = None,
        writers: list[str] | None = None,
    ):
        if self.collection_id:
            # We've already been uploaded!
            return self.colleciton_id

        if not skip_preflight:
            # This runs _all_ preflight checks - for all connected collecitons and products.
            self.preflight()

        self.collection_id = collections.create(
            client=client,
            name=self.name,
            description=self.description,
            readers=readers,
            writers=writers,
            console=console,
        )

        product_ids_to_connect = [
            x._upload(
                client=client,
                console=console,
                readers=readers,
                writers=writers,
                skip_preflight=True,
            )
            for x in self.products
        ]

        for product_id in product_ids_to_connect:
            collections.add(
                client=client,
                id=self.collection_id,
                product=product_id,
                console=console,
            )

        # Recurse!
        child_collection_ids_to_connect = [
            x._upload(
                client=client,
                console=console,
                skip_preflight=True,
                readers=readers,
                writers=writers,
            )
            for x in self.collections
        ]

        for collection_id in child_collection_ids_to_connect:
            relationships.add_child_collection(
                client=client,
                parent=self.collection_id,
                child=collection_id,
                console=console,
            )

        return self.collection_id


class RemoteCollection(CollectionInstance):
    """
    A remote collection, created by pushing from HIPPO.
    """

    collection_id: str
    name: str
    description: str
    products: list[ProductInstance] = []
    collections: list[CollectionInstance] = []
    owner: str
    readers: list[str] = []
    writers: list[str] = []

    original_name: str
    original_description: str
    original_product_ids: set[str] = Field(default_factory=set)
    original_collection_ids: set[str] = Field(default_factory=set)

    @classmethod
    def pull(
        cls,
        collection_id: str,
        client: httpx.Client,
        cache: MultiCache,
        console: Console,
        pull_children: bool = True,
        realize_sources: bool = True,
    ) -> "RemoteCollection":
        collection = collections.read(client=client, id=collection_id, console=console)

        if realize_sources:
            # Ensure it's cached
            collections.cache(
                client=client,
                cache=cache,
                id=collection_id,
                console=console,
            )

        if pull_children:
            products = [
                RemoteProduct.pull(
                    product_id=str(x.id),
                    client=client,
                    cache=cache,
                    console=console,
                    realize_sources=realize_sources,
                )
                for x in collection.products
            ]
            child_collections = [
                RemoteCollection.pull(
                    collection_id=str(x.id),
                    client=client,
                    cache=cache,
                    console=console,
                    realize_sources=realize_sources,
                    pull_children=pull_children,
                )
                for x in collection.child_collections
            ]
        else:
            products = []
            child_collections = []

        return cls(
            collection_id=str(collection.id),
            name=collection.name,
            description=collection.description,
            products=products,
            owner=collection.owner,
            collections=child_collections,
            readers=collection.readers,
            writers=collection.writers,
            original_name=collection.name,
            original_description=collection.description,
            original_product_ids={str(p.id) for p in collection.products},
            original_collection_ids={str(c.id) for c in collection.child_collections},
        )

    @classmethod
    def read(cls, directory: Path | str, allow_incomplete: bool) -> "RemoteProduct":
        """
        Read a collection that was serialized to disk, usually with the
        `henry collection download $ID` command.
        """
        directory = Path(directory)

        with open(directory / "collection.json", "r") as handle:
            core_metadata = ReadCollectionResponse.model_validate_json(handle.read())

        products = []

        for product in core_metadata.products:
            try:
                products.append(
                    RemoteProduct.read(
                        directory / product.name, allow_incomplete=allow_incomplete
                    )
                )
            except FileNotFoundError:
                if not allow_incomplete:
                    raise CollectionIncompleteError(
                        f"Product {product.name} not found, consider re-downloading "
                        "or setting allow_incomplete=True"
                    )
            except ProductIncompleteError as e:
                if not allow_incomplete:
                    raise CollectionIncompleteError(
                        f"Collection not complete due to error reading {product.name}: {e}"
                    )

        collections = []

        for collection in core_metadata.child_collections:
            try:
                collections.append(
                    RemoteCollection.read(
                        directory / collection.name, allow_incomplete=allow_incomplete
                    )
                )
            except FileNotFoundError:
                if not allow_incomplete:
                    raise CollectionIncompleteError(
                        f"Child collection {collection.name} not found, consider "
                        "re-downloading or setting allow_incomplete=True"
                    )
            except CollectionIncompleteError as e:
                if not allow_incomplete:
                    raise e

        return cls(
            collection_id=str(core_metadata.id),
            name=core_metadata.name,
            description=core_metadata.description,
            products=products,
            owner=core_metadata.owner,
            collections=collections,
            readers=core_metadata.readers,
            writers=core_metadata.writers,
            original_name=core_metadata.name,
            original_description=core_metadata.description,
            original_product_ids={str(p.id) for p in core_metadata.products},
            original_collection_ids={
                str(c.id) for c in core_metadata.child_collections
            },
        )

    def __get_global_index(self, key: int, /) -> tuple[str, int]:
        if key < 0:
            return self.__get_global_index(self.__len__() + key)
        elif key < len(self.products):
            return "p", key
        elif key < self.__len__():
            return "c", key - len(self.products)
        else:
            raise IndexError(f"Index {key} invalid for object of size {self.__len__()}")

    def __repr__(self):
        return (
            f"RemoteCollection(name='{self.name}', description='{self.description}', "
            f"products=[{', '.join([x.__repr__() for x in self.products])}], "
            f"collections=[{', '.join([x.__repr__() for x in self.collections])}])"
        )

    def __str__(self):
        return (
            f"{self.name} ({self.description}) collection with {len(self.products)} "
            f"products and {len(self.collections)} child collections"
        )

    def __len__(self) -> int:
        return len(self.products) + len(self.collections)

    def __getitem__(self, key: int, /) -> CollectionInstance | ProductInstance:
        arr, index = self.__get_global_index(key)
        if arr == "p":
            return self.products[index]
        else:
            return self.collections[index]

    def __iter__(self) -> Iterable[CollectionInstance | ProductInstance]:
        return chain(self.products, self.collections)

    def __reversed__(self) -> Iterable[ProductInstance | CollectionInstance]:
        return chain(self.collections.__reversed__(), self.products.__reversed__())

    def __contains__(self, item: CollectionInstance | ProductInstance, /):
        match item:
            case CollectionInstance():
                self.collections.__contains__(item)
            case ProductInstance():
                self.products.__contains__(item)
            case _:
                raise TypeError(
                    f"Only collections or products are members of collections, found {type(item)}"
                )

    def append(self, value: CollectionInstance | ProductInstance, /):
        match value:
            case CollectionInstance():
                self.collections.append(value)
            case ProductInstance():
                self.products.append(value)
            case _:
                raise TypeError(
                    f"Only collections or products may be appended to a collection, found {type(value)}"
                )

    def extend(self, iterable: Iterable[CollectionInstance | ProductInstance], /):
        for a in iterable:
            self.append(a)

    def pop(self, key: int, /):
        raise RuntimeError(
            "Cannot remove items from a remote collection. If you need a custom "
            "list, create a LocalCollection with what you need instead"
        )

    def remove(self, value, /):
        raise RuntimeError(
            "Cannot remove items from a remote collection. If you need a custom "
            "list, create a LocalCollection with what you need instead"
        )

    def preflight(self):
        """
        Run a pre-flight check on this collection. This checks that:

        a) The collection has a valid name and description.
        b) Runs preflight checks for all connected products.
        c) Runs preflight checks for all connected collections.
        """

        if self.collection_id:
            # We've already flown!
            return

        for preflightable in self:
            preflightable.preflight()

        if len(self.name) < 2 or not isinstance(self.name, str):
            raise PreflightFailedError(
                f"Name: {self.name} is not a valid name; ensure it is at least 2 characters and a valid string"
            )

        if len(self.description) < 2 or not isinstance(self.description, str):
            raise PreflightFailedError(
                f"Description: {self.description} is not a valid description; ensure it is at least 2 characters and a valid string"
            )

    def _upload(
        self,
        client: httpx.Client,
        console: Console,
        skip_preflight: bool = False,
        readers: list[str] | None = None,
        writers: list[str] | None = None,
    ):
        if self.collection_id:
            # We've already been uploaded!
            return self.colleciton_id

        if not skip_preflight:
            # This runs _all_ preflight checks - for all connected collecitons and products.
            self.preflight()

        for product in self.products:
            product._upload(
                client=client,
                console=console,
                skip_preflight=skip_preflight,
                readers=readers,
                writers=writers,
            )

        product_ids_to_connect = [
            p.id for p in self.products if p.id not in self.original_product_ids
        ]

        for product_id in product_ids_to_connect:
            collections.add(
                client=client,
                id=self.collection_id,
                product=product_id,
                console=console,
            )

        for collection in self.collections:
            collection._upload(
                client=client,
                console=console,
                skip_preflight=skip_preflight,
                readers=readers,
                writers=writers,
            )

        # Recurse!
        child_collection_ids_to_connect = [
            c.id for c in self.collections if c.id not in self.original_collection_ids
        ]

        for collection_id in child_collection_ids_to_connect:
            relationships.add_child_collection(
                client=client,
                parent=self.collection_id,
                child=collection_id,
                console=console,
            )

        return self.collection_id



"""
The main FastAPI endpoints.
"""

from contextlib import asynccontextmanager

from beanie import init_beanie
from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import HTMLResponse, RedirectResponse
from loguru import logger
from pymongo import AsyncMongoClient
from starlette.exceptions import HTTPException
from starlette.middleware.cors import CORSMiddleware

from hipposerve.api.product import product_router
from hipposerve.api.relationships import relationship_router
from hipposerve.api.soauth import setup_auth
from hipposerve.database import BEANIE_MODELS
from hipposerve.service.collection import CollectionNotFound
from hipposerve.service.product import ProductNotFound
from hipposerve.settings import SETTINGS
from hipposerve.storage import Storage
from hipposerve.web.auth import UnauthorizedException
from hipposerve.web.router import templates


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Initialize application services."""
    app.db = AsyncMongoClient(SETTINGS.mongo_uri).account
    await init_beanie(app.db, document_models=BEANIE_MODELS)
    app.storage = Storage(
        url=SETTINGS.minio_url,
        presign_url=SETTINGS.minio_presign_url,
        upgrade_presign_url_to_https=SETTINGS.minio_upgrade_presign_url_to_https,
        secure=SETTINGS.minio_secure,
        cert_check=SETTINGS.minio_cert_check,
        access_key=SETTINGS.minio_access,
        secret_key=SETTINGS.minio_secret,
    )

    logger.info("Startup complete")
    yield
    logger.info("Shutdown complete")


app = setup_auth(
    FastAPI(
        title=SETTINGS.title,
        description=SETTINGS.description,
        lifespan=lifespan,
        docs_url="/docs" if SETTINGS.debug else None,
        redoc_url="/redoc" if SETTINGS.debug else None,
    )
)

# Routers
app.include_router(product_router)
app.include_router(relationship_router)

if SETTINGS.web:  # pragma: no cover
    from hipposerve.web import static_files, web_router

    logger.info("Web interface enabled, serving it from {}", web_router.prefix)

    app.include_router(web_router)
    app.mount(**static_files)

    @app.exception_handler(UnauthorizedException)
    async def unauthorized_exception_handler(
        request: Request, exc: UnauthorizedException
    ):
        if request.user.is_authenticated:
            return HTMLResponse(
                content="You do not have permission to access this resource",
                status_code=403,
            )
        else:
            return RedirectResponse(app.login_url)

    def not_found_template(
        request: Request,
        type: str,
        requested_id: str | None,
    ):
        error_details = {
            "type": type,
        }

        if error_details["type"] != "generic":
            error_details["requested_id"] = requested_id

        return templates.TemplateResponse(
            "404.html",
            {
                "request": request,
                "error_details": error_details,
                "user": request.user.display_name,
                "web_root": SETTINGS.web_root,
            },
            status_code=404,
        )

    if not SETTINGS.debug:
        logger.warning(
            "Debug mode is disabled, enabling wide-ranging HTTPException handler"
        )

        @app.exception_handler(HTTPException)
        async def page_not_found_handler(
            request: Request,
            exc: HTTPException,
        ):
            return not_found_template(request, "generic", None)

        logger.warning(
            "Debug mode is disabled, enabling wide-ranging validation handler"
        )

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(
            request: Request,
            exc: RequestValidationError,
        ):
            requested_item_type = "generic"
            requested_id = None
            try:
                requested_id = request.path_params["id"]
                if request.url.path.find("collections"):
                    requested_item_type = "collection"
                elif request.url.path.find("products"):
                    requested_item_type = "product"
                return not_found_template(request, requested_item_type, requested_id)
            except KeyError:
                return not_found_template(request, requested_item_type, requested_id)

    @app.exception_handler(CollectionNotFound)
    async def collection_not_found_handler(
        request: Request,
        exc: CollectionNotFound,
    ):
        requested_id = request.path_params["id"]
        return not_found_template(request, "collection", requested_id)

    @app.exception_handler(ProductNotFound)
    async def product_not_found_handler(
        request: Request,
        exc: ProductNotFound,
    ):
        requested_id = request.path_params["id"]
        return not_found_template(request, "product", requested_id)

    @app.get("/")
    async def web_redirect(request: Request):
        return RedirectResponse(SETTINGS.web_root)


if SETTINGS.add_cors:
    logger.warning(
        "Completely open CORS policy enabled, designed for testing and development"
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )



"""
Request and response models for the groups endpoints.
"""

from pydantic import BaseModel

from hipposerve.service.users import Privilege


class CreateGroupRequest(BaseModel):
    privileges: list[Privilege]
    description: str | None


class CreateGroupResponse(BaseModel):
    group_name: str


class ReadGroupResponse(BaseModel):
    name: str
    privileges: list[Privilege]
    description: str | None


class UpdateGroupAccessRequest(BaseModel):
    add_access_control: list[Privilege] | None = None
    remove_access_control: list[Privilege] | None = None


class UpdateGroupAccessResponse(BaseModel):
    access_controls: list[Privilege]


class UpdateGroupDescriptionRequest(BaseModel):
    description: str | None = None


class UpdateGroupDescriptionResponse(BaseModel):
    description: str



from typing import Literal

from pydantic import Field

from .base import BaseMetadata


class NumericMetadata(BaseMetadata):
    """
    A simple metadata type for a numeric value.
    """

    metadata_type: Literal["numeric"] = "numeric"

    value: float = Field(json_schema_extra=dict(min=-1e100, max=1e100))



"""
Helpers for rendering cli elements.
"""

import rich

from hippoclient.caching import MultiCache
from hipposerve.api.models.relationships import (
    ReadCollectionProductResponse,
    ReadCollectionResponse,
)
from hipposerve.database import FileMetadata, ProductMetadata


def render_version_list(
    versions: list[str], current_version: str, requested_version: str
) -> str:
    """
    Render a list of versions to a bbcode string.

    Current version is bolded, requested version is green and underlined.
    """

    output = ""

    for version in versions:
        start = ""
        end = ""
        if version == current_version:
            start = f"[b]{start}"
            end = f"{end}[/b]"
        if version == requested_version:
            start = f"[u][color=green]{start}"
            end = f"{end}[/color][/u]"

        output += f"{start}{version}{end} "

    return output


def render_product_metadata_list(
    input: list[ProductMetadata | ReadCollectionProductResponse],
) -> rich.table.Table:
    """
    Render a list of product meadata into a rich table.
    """

    table = rich.table.Table(title="Products")

    table.add_column("ID", justify="center", width=24)
    table.add_column("Name", justify="left")
    table.add_column("Version", justify="center", width=7)
    table.add_column("Uploaded", justify="center", width=16)

    for product in input:
        table.add_row(
            str(product.id),
            product.name,
            product.version,
            product.uploaded.strftime("%Y-%m-%d %H:%M"),
        )

    return table


def render_collection_metadata_list(
    input: list[ReadCollectionResponse],
) -> rich.table.Table:
    """
    Render a list of collection metadata into a rich table.
    """

    table = rich.table.Table(title="Collections")

    table.add_column("ID", justify="center", width=24)
    table.add_column("Name", justify="left")
    table.add_column("Description", justify="left")

    for collection in input:
        description = collection.description.strip("\n")

        # Truncate it if it's too long
        if len(description) > 512:
            description = description[:512] + "..."

        table.add_row(str(collection.id), collection.name, description)

    return table


def render_source_list(
    input: dict[str, FileMetadata], cache: MultiCache | None = None
) -> rich.table.Table:
    """
    Render a list of source metadata into a rich table. If you provide
    a cache, we also query it to see if the files are cached.
    """

    table = rich.table.Table(
        "Name", "Description", "UUID", "Size [B]", "Cached", title="Sources"
    )

    for slug, source in input.items():
        if cache is not None:
            try:
                cache.available(source.uuid)
                cached = "Yes"
            except FileNotFoundError:
                cached = "No"
        else:
            cached = "Unknown"

        table.add_row(
            slug,
            source.name,
            source.description or "",
            source.uuid,
            str(source.size),
            cached,
        )

    return table



"""
Base metadata type that all metadata must inherit from
"""

from typing import ClassVar

from pydantic import BaseModel


class BaseMetadata(BaseModel):
    metadata_type: str
    valid_slugs: ClassVar[set[str]] = {"data"}



"""
A simple python wrapper for the HTTP for hippo.

API key can be stored in the hippo_API_KEY environment variable.
Host can be stored in the hippo_HOST environment variable.
"""

__all__ = [
    "Client",
    "collections",
    "product",
    "relationships",
    "caching",
]

from . import caching, collections, product, relationships
from .core import Client



"""
The caching layer for sources used by the hippo client.

A cache is:

a) A directory on your filesystem
b) A set of files in that directory
c) A SQLite database that stores metadata about the files
   that are being cached.

Because marshalling is so simple here we do not use
SQLAlchemy or any other ORM. We just use the built-in
Python sqlite3 module.
"""

import os
import shutil
import sqlite3
from pathlib import Path

from pydantic import BaseModel
from rich.console import Console

from hippoclient.downloads import downloader
from hipposerve.database import FileMetadata


class CacheNotWriteableError(Exception):
    """
    Raised when the cache is not writeable.
    """

    def __init__(self):
        super().__init__("The cache is not writeable")


class Cache(BaseModel):
    """
    A cache for sources used by the hippo client. All sources are labelled by their
    unique UUID, given by the main mongodb database (note this is not _id).

    The general execution flow when using a (set of) cache(s) is:

    1. Check if the file is in the cache with ``available``. If it is, you will be
       immediately given the on-disk path to the file. If not, we will raise a
       FileNotFoundError.
    2. Check the other caches if it is not available.
    3. Attain a pre-signed URL for your source.
    4. Fetch the source from the pre-signed URL with your preferred cache with ``get``.
       This will return you an on-disk path to the downloaded source within your cache
       area. You can check which caches are writeable with ``writeable``.

    So for a list of caches, you would do something like:

    .. code-block::python

        MY_FILE_ID = "123456789012345678901234"
        path: Path | None = None

        for cache in caches:
            try:
                path = cache.available(MY_FILE_ID)
                break
            except FileNotFoundError:
                continue

        if path is None:
            for cache in filter(lambda c: c.writeable, caches):
                try:
                    path = cache.get(
                        id=MY_FILE_ID,
                        path="path/to/file",
                        checksum="checksum",
                        size=1234,
                        presigned_url="https://example.com/presigned-url",
                    )
                    break
                except {SOME EXCEPTION}:
                    continue

        if path is None:
            raise FileNotFoundError

    There are a number of utilities implemented in the MultiCache object that perform
    these kind of duties for you.
    """

    path: Path
    database_name: str = "cache.db"

    _database: Path
    _connection: sqlite3.Connection

    def model_post_init(self, __context):
        self._database = self.path / self.database_name
        self._connection = self._initialize_database()

    def _initialize_database(self) -> sqlite3.Connection:
        """
        Initialize the cache database.
        """

        if self._database.exists():
            return sqlite3.connect(self._database)
        else:
            # Create the database
            connection = sqlite3.connect(self._database)

            cursor = connection.cursor()
            cursor.execute(
                "CREATE TABLE sources (id PRIMARY KEY, path, checksum, size, available)"
            )
            connection.commit()

            return connection

    @property
    def writeable(self) -> bool:
        """
        Check if the cache is writeable.
        """

        return os.access(self._database, os.W_OK)

    def _add(self, id: str, path: str, checksum: str, size: int):
        """
        Add a source to the cache. Initially marks it as unavailable. This should
        be called _before_ downloading the actual source.
        """

        cursor = self._connection.cursor()
        cursor.execute(
            "INSERT INTO sources (id, path, checksum, size, available) VALUES (?, ?, ?, ?, ?)",
            (str(id), str(path), str(checksum), int(size), False),
        )
        self._connection.commit()

    def _mark_available(self, id: str):
        """
        Mark a source as available.
        """

        cursor = self._connection.cursor()
        cursor.execute(
            "UPDATE sources SET available = ? WHERE id = ?",
            (True, str(id)),
        )
        self._connection.commit()

    def _mark_unavailable(self, id: str):
        """
        Mark a source as unavailable.
        """

        cursor = self._connection.cursor()
        cursor.execute(
            "UPDATE sources SET available = ? WHERE id = ?",
            (False, str(id)),
        )
        self._connection.commit()

    def _get(self, id: str) -> Path | None:
        """
        Get a source from the cache. Or, if it's not available, return None.
        If it is available we return the absolute path on the system to the file.
        """

        cursor = self._connection.cursor()

        cursor.execute(
            "SELECT path FROM sources WHERE id = ? AND available = ?",
            (str(id), True),
        )

        result = cursor.fetchone()

        if result is None:
            return None
        else:
            return self.path / Path(result[0])

    def _remove(self, id: str):
        """
        Remove a source from the cache.
        """

        # Get it first, if it's there then we need to delete it.
        path = self._get(id)

        if path is not None:
            path.unlink()

        cursor = self._connection.cursor()
        cursor.execute(
            "DELETE FROM sources WHERE id = ?",
            (str(id),),
        )
        self._connection.commit()

    def _fetch(
        self, id: str, path: str, checksum: str, size: int, presigned_url: str
    ) -> Path:
        """
        Fetch a source from the presigned URL we are provided, store it in
        the cache, and return the path to the file.

        Parameters
        ----------
        id : str
            The ID of the source
        path : str
            The path to the source
        checksum : str
            The checksum of the source
        size : int
            The size of the source
        presigned_url : str
            The presigned URL to fetch the source from

        Returns
        -------
        Path
            The path to the source in the cache

        Raises
        ------
        CacheNotWriteableError
            If the cache is not writeable
        """

        if not self.writeable:
            raise CacheNotWriteableError

        try:
            self._add(id=id, path=path, checksum=checksum, size=size)

            # Download the file
            destination_path = self.path / Path(path)
            destination_path.parent.mkdir(parents=True, exist_ok=True)

            downloader(
                presigned_url=presigned_url,
                output_destination=destination_path,
                console=Console(),
            )

            self._mark_available(id)
        except BaseException as e:
            # Atomic! Not super threadsafe but this is sqlite anyway. Usually when someone kills this
            # with a KeyboardInterrupt -- we don't wanna leave dangling db items. We could probably handle
            # this with an actual transaction...
            self._remove(id=id)
            try:
                shutil.rmtree(destination_path)
            except NotADirectoryError:
                os.remove(destination_path)
            except FileNotFoundError:
                pass
            raise e

        return destination_path

    def get(
        self, id: str, path: str, checksum: str, size: int, presigned_url: str
    ) -> Path:
        """
        Get a source from the cache. If it's not available, fetch it from the
        presigned URL and store it in the cache.

        Parameters
        ----------
        id : str
            The ID of the source
        path : str
            The path to the source
        checksum : str
            The checksum of the source
        size : int
            The size of the source
        presigned_url : str
            The presigned URL to fetch the source from, if it not in the cache.
        """

        cached = self._get(id)

        if cached is not None:
            return cached
        else:
            return self._fetch(
                id=id,
                path=path,
                checksum=checksum,
                size=size,
                presigned_url=presigned_url,
            )

    def available(self, id: str) -> Path:
        """
        Check if a source is available in the cache with id ``id``.
        If it is, we return the Path. If not, we raise a FileNotFoundError.

        Parameters
        ----------
        id : str
            The ID of the source

        Returns
        -------
        Path
            The path to the source in the cache

        Raises
        ------
        FileNotFoundError
            If the source is not available in the cache
        """

        path = self._get(id)

        if path is None:
            raise FileNotFoundError

        return path

    @property
    def complete_id_list(self) -> list[str]:
        """
        List all the IDs in the cache
        """

        cursor = self._connection.cursor()
        cursor.execute("SELECT id FROM sources")

        return [row[0] for row in cursor.fetchall()]


class MultiCache(BaseModel):
    """
    An interface to multiple caches. There is one main rule:

    Caches are ordered by priority. We will always try the first caches in
    the list first, and files will be ingested on a first-cache-first-served
    basis.
    """

    caches: list[Cache]

    def available(self, id: str) -> Path:
        """
        Check if a source is available in any cache with id ``id``.
        If it is, we return the Path. If not, we raise a FileNotFoundError.

        Parameters
        ----------
        id : str
            The ID of the source

        Returns
        -------
        Path
            The path to the source in the cache

        Raises
        ------
        FileNotFoundError
            If the source is not available in the cache
        """

        for cache in self.caches:
            try:
                return cache.available(id)
            except FileNotFoundError:
                continue

        raise FileNotFoundError

    def get(self, id: str, path: str, checksum: str, size: int, presigned_url: str):
        """
        Get and store an item in the cache. You should likely use this only after
        ``available`` has raised a FileNotFoundError.

        Parameters
        ----------
        id : str
            The ID of the source
        path : str
            The path to the source
        checksum : str
            The checksum of the source
        size : int
            The size of the source
        presigned_url : str
            The presigned URL to fetch the source from, if it not in the cache.

        Returns
        -------
        Path
            The path to the source in the cache

        Raises
        ------
        CacheNotWriteableError
            If no caches are writeable
        """

        for cache in self.caches:
            if cache.writeable:
                return cache.get(
                    id=id,
                    path=path,
                    checksum=checksum,
                    size=size,
                    presigned_url=presigned_url,
                )

        raise CacheNotWriteableError

    def remove(self, id: str):
        """
        Remove a source from all caches.
        """

        for cache in self.caches:
            cache._remove(id)

    def names_to_paths(self, file_metadata: list[FileMetadata]) -> dict[str, Path]:
        """
        Convert a list of FileMetadata objects to a dictionary of names to paths.
        """

        paths = {}

        for cache in self.caches:
            for file in file_metadata:
                path = cache.available(file.uuid)
                paths[file.name] = path

        return paths

    def uuids_to_paths(self, file_metadata: list[FileMetadata]) -> dict[str, Path]:
        """
        Convert a list of FileMetadata objects to a dictionary of UUIDs to paths.
        """

        paths = {}

        for cache in self.caches:
            for file in file_metadata:
                path = cache.available(file.uuid)
                paths[file.uuid] = path

        return paths


def clear_all(cache: Cache):
    """
    Clear all the caches.
    """

    for file in cache.complete_id_list:
        cache._remove(file)


def clear_single(cache: Cache, id: str):
    """
    Clear a single cached item.
    """

    cache._remove(id)



"""
Routes for the product service.
"""

import asyncio

from beanie import PydanticObjectId
from fastapi import APIRouter, HTTPException, Request, status
from fastapi.responses import RedirectResponse
from loguru import logger

from hipposerve.api.models.product import (
    CompleteProductRequest,
    CreateProductRequest,
    CreateProductResponse,
    ReadFilesResponse,
    ReadProductResponse,
    UpdateProductRequest,
    UpdateProductResponse,
)
from hipposerve.database import ProductMetadata
from hipposerve.service import acl, product, storage, users
from hipposerve.service.auth import AuthenticationError, requires

product_router = APIRouter(prefix="/product")

DEFAULT_USER_USER_NAME = "default_user"


@product_router.put("/new")
@requires(["hippo:admin", "hippo:write"])
async def create_product(
    request: Request,
    model: CreateProductRequest,
) -> CreateProductResponse:
    """
    Create a new product, returning the pre-signed URLs for the sources.
    """

    logger.info(
        "Create product request: {} from {}", model.name, request.user.display_name
    )

    if await product.exists(name=model.name):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT, detail="Product already exists"
        )

    item, presigned = await product.create(
        name=model.name,
        description=model.description,
        metadata=model.metadata,
        sources=model.sources,
        user_name=request.user.display_name,
        storage=request.app.storage,
        product_readers=model.product_readers,
        product_writers=model.product_writers,
        mutlipart_size=model.multipart_batch_size,
    )

    logger.info(
        "Successfully created {} pre-signed URL(s) for product upload {} (id: {}) from {}",
        sum(len(x) for x in presigned.values()),
        model.name,
        item.id,
        request.user.display_name,
    )

    return CreateProductResponse(id=item.id, upload_urls=presigned)


@product_router.get("/search/{text}")
@requires(["hippo:admin", "hippo:read"])
async def search(
    text: str,
    request: Request,
) -> list[ProductMetadata]:
    """
    Search for a product by name.
    """

    logger.info(
        "Search for product {} request from {}", text, request.user.display_name
    )

    items = await product.search_by_name(
        name=text, groups=request.user.groups, scopes=request.auth.scopes
    )

    logger.info(
        "Successfully found {} product(s) matching {} requested by {}",
        len(items),
        text,
        request.user.display_name,
    )

    return await asyncio.gather(*[item.to_metadata() for item in items])


@product_router.post("/{id}/complete")
@requires(["hippo:admin", "hippo:write"])
async def complete_product(
    id: PydanticObjectId,
    request: Request,
    model: CompleteProductRequest,
) -> None:
    """
    Complete a product's upload. Must be called before the sources are available.
    """
    logger.info(
        "Complete product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    success = await product.complete(
        product=item,
        storage=request.app.storage,
        headers=model.headers,
        sizes=model.sizes,
    )

    if not success:
        raise HTTPException(
            status_code=status.HTTP_424_FAILED_DEPENDENCY,
            detail="Not all sources were present.",
        )

    logger.info("Successfully completed product {} (id: {})", item.name, item.id)


@product_router.get("/{id}")
@requires(["hippo:admin", "hippo:read"])
async def read_product(
    id: PydanticObjectId,
    request: Request,
) -> ReadProductResponse:
    """
    Read a single product's metadata.
    """

    logger.info("Read product request for {} from {}", id, request.user.display_name)

    try:
        product_item = await product.read_by_id(
            id, request.user.groups, scopes=request.auth.scopes
        )
        item = await product_item.to_metadata()

        response = ReadProductResponse(
            current_present=item.current,
            current=item.version if item.current else None,
            requested=item.version,
            versions={item.version: item},
        )

        logger.info(
            "Successfully read product {} (id: {}) requested by {}",
            item.name,
            item.id,
            request.user.display_name,
        )

        return response
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )


@product_router.get("/{id}/tree")
@requires(["hippo:admin", "hippo:read"])
async def read_tree(
    id: PydanticObjectId,
    request: Request,
) -> ReadProductResponse:
    """
    Read a single product's entire history.
    """

    logger.info(
        "Read product tree request for {} from {}", id, request.user.display_name
    )

    try:
        requested_item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        current_item = await product.walk_to_current(
            product=requested_item,
            groups=request.user.groups,
            scopes=request.auth.scopes,
        )
        history = await product.walk_history(product=current_item)

        if not current_item.current:
            raise HTTPException(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Unable to find the current version of the requested item",
            )

        response = ReadProductResponse(
            current_present=True,
            current=current_item.version,
            requested=requested_item.version,
            versions=history,
        )

        logger.info(
            "Successfully read product tree for {} (id: {}) requested by {}",
            requested_item.name,
            requested_item.id,
            request.user.display_name,
        )

        return response
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )


@product_router.get("/{id}/files")
@requires(["hippo:admin", "hippo:read"])
async def read_files(id: PydanticObjectId, request: Request) -> ReadFilesResponse:
    """
    Read a single product's including pre-signed URLs for downloads.
    """

    logger.info("Read files request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    files = await product.read_files(product=item, storage=request.app.storage)

    logger.info(
        "Read {} pre-signed URLs for product {} (id: {}) requested by {}",
        len(files),
        item.name,
        item.id,
        request.user.display_name,
    )

    return ReadFilesResponse(
        product=await item.to_metadata(),
        files=files,
    )


@product_router.get("/{id}/{slug}")
@requires(["hippo:admin", "hippo:read"])
async def read_slug(request: Request, id: str, slug: str) -> RedirectResponse:
    NOT_FOUND = HTTPException(
        status_code=status.HTTP_404_NOT_FOUND,
        detail=f"Slug {slug} not found for product {id}",
    )
    try:
        product_instance = await product.read_by_id(
            id, request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise NOT_FOUND

    # If they have not 401/403'd from reading the product, they can read sources
    try:
        presigned = await storage.read(
            file=product_instance.sources[slug], storage=request.app.storage
        )
    except (FileNotFoundError, KeyError):
        raise NOT_FOUND

    return RedirectResponse(url=presigned, status_code=status.HTTP_302_FOUND)


@product_router.post("/{id}/diff")
@requires(["hippo:admin", "hippo:write"])
async def metadata_diff(
    id: PydanticObjectId,
    model: UpdateProductRequest,
    request: Request,
):
    """
    Calculates a diff between the current product and any new metadata.
    """

    logger.info("Update product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    try:
        acl.check_user_access(
            user_groups=request.user.groups,
            document_groups=item.writers,
            scopes=request.auth.scopes,
        )
    except AuthenticationError:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="User does not have write access to this product",
        )

    # Check the _new_ owner is an existing user
    if model.owner is not None:
        try:
            await users.confirm_user(name=model.owner)
        except users.UserNotFound:
            raise HTTPException(
                status_code=status.HTTP_406_NOT_ACCEPTABLE, detail="User not found"
            )

    if not item.current:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Cannot update a product that is not the current version",
        )

    diff = await product.metadata_diff(
        product=item,
        name=model.name,
        description=model.description,
        metadata=model.metadata,
        level=model.level,
    )

    logger.debug(
        "Found metadata diff for product {} (id: {}): {}", item.name, item.id, diff
    )

    return diff


@product_router.post("/{id}/update")
@requires(["hippo:admin", "hippo:write"])
async def update_product(
    id: PydanticObjectId,
    model: UpdateProductRequest,
    request: Request,
) -> UpdateProductResponse:
    """
    Update a product's details.
    """

    logger.info("Update product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found"
        )

    try:
        acl.check_user_access(
            user_groups=request.user.groups,
            document_groups=item.writers,
            scopes=request.auth.scopes,
        )
    except AuthenticationError:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="User does not have write access to this product",
        )

    # Check the _new_ owner is an existing user
    if model.owner is not None:
        try:
            await users.confirm_user(name=model.owner)
        except users.UserNotFound:
            raise HTTPException(
                status_code=status.HTTP_406_NOT_ACCEPTABLE, detail="User not found."
            )

    if model.level:
        try:
            new_product, upload_urls = await product.update(
                product=item,
                access_groups=request.user.groups,
                scopes=request.auth.scopes,
                name=model.name,
                description=model.description,
                metadata=model.metadata,
                new_sources=model.new_sources,
                replace_sources=model.replace_sources,
                drop_sources=model.drop_sources,
                storage=request.app.storage,
                level=model.level,
            )
        except AuthenticationError:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="User does not have write access to this product",
            )

        logger.info(
            "Successfully updated product {} (new id: {}; {}; old id: {}; {}) from {}",
            new_product.name,
            new_product.id,
            new_product.version,
            item.id,
            item.version,
            request.user.display_name,
        )
    else:
        new_product = item
        upload_urls = {}

    new_product = await acl.update_access_control(
        doc=new_product,
        owner=model.owner,
        add_readers=model.add_readers,
        remove_readers=model.remove_readers,
        add_writers=model.add_writers,
        remove_writers=model.remove_writers,
    )

    logger.info(
        "Successfully updated product {} (id: {}) with new access control policy",
        new_product.name,
        new_product.id,
    )

    return UpdateProductResponse(
        version=new_product.version,
        id=new_product.id,
        upload_urls=upload_urls,
    )


@product_router.post("/{id}/confirm")
@requires(["hippo:admin", "hippo:read"])
async def confirm_product(id: PydanticObjectId, request: Request) -> None:
    """
    Confirm a product's sources.
    """

    logger.info("Confirm product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
        success = await product.confirm(
            product=item,
            storage=request.app.storage,
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    if not success:
        raise HTTPException(
            status_code=status.HTTP_424_FAILED_DEPENDENCY,
            detail="Not all sources were present.",
        )

    logger.info("Successfully confirmed product {} (id: {})", item.name, item.id)


@product_router.post("/{id}/pin")
@requires(["hippo:admin"])
async def pin_product(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Pin a product to place it on the homepage.
    """

    logger.info("Pin product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    item.pinned = True
    await item.save()

    logger.info("Successfully pinned product {} (id: {})", item.name, item.id)


@product_router.post("/{id}/unpin")
@requires(["hippo:admin"])
async def unpin_product(
    id: PydanticObjectId,
    request: Request,
) -> None:
    """
    Unpin a product to remvove it from the homepage.
    """

    logger.info("Unpin product request for {} from {}", id, request.user.display_name)

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    item.pinned = False
    await item.save()

    logger.info("Successfully unpinned product {} (id: {})", item.name, item.id)


@product_router.delete("/{id}")
@requires(["hippo:admin", "hippo:write"])
async def delete_product(
    id: PydanticObjectId,
    request: Request,
    data: bool = False,
) -> None:
    """
    Delete a product.
    """

    logger.info(
        "Delete (single) product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    await product.delete_one(
        product=item,
        access_groups=request.user.groups,
        storage=request.app.storage,
        data=data,
        scopes=request.auth.scopes,
    )

    logger.info(
        "Successfully deleted product {} (id: {}) including data"
        if data
        else "Successfully deleted product {} (id: {}) excluding data",
        item.name,
        item.id,
    )


@product_router.delete("/{id}/tree")
@requires(["hippo:admin", "hippo:write"])
async def delete_tree(
    id: PydanticObjectId,
    request: Request,
    data: bool = False,
) -> None:
    """
    Delete a product.
    """

    logger.info(
        "Delete (tree) product request for {} from {}", id, request.user.display_name
    )

    try:
        item = await product.read_by_id(
            id=id, groups=request.user.groups, scopes=request.auth.scopes
        )
    except product.ProductNotFound:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Product not found."
        )

    await product.delete_tree(
        product=item,
        access_groups=request.user.groups,
        storage=request.app.storage,
        data=data,
        scopes=request.auth.scopes,
    )

    logger.info(
        "Successfully deleted product tree for {} (id: {}) including data"
        if data
        else "Successfully deleted product tree for {} (id: {}) excluding data",
        item.name,
        item.id,
    )



from urllib.parse import urlencode

from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from loguru import logger
from starlette.middleware.base import BaseHTTPMiddleware

from hipposerve.service import users as user_service
from hipposerve.settings import SETTINGS


class UserMiddleware(BaseHTTPMiddleware):
    """
    Middleware to handle user authentication and updates. Responsible for keeping
    the in-database user record up to date with the one provided by SOAuth.

    This middleware checks if the user is authenticated and updates their last
    access time. If the user does not exist, it creates a new user record.
    """

    async def dispatch(self, request: Request, call_next):
        user_obj = request.scope.get("user")
        if user_obj and user_obj.is_authenticated:
            try:
                # Do not need to check if the user exists as touching
                # does that for us.
                await user_service.touch_last_access_time(
                    name=request.user.display_name
                )
            except user_service.UserNotFound:
                await user_service.create(
                    name=user_obj.display_name, email=user_obj.email
                )

        return await call_next(request)


def setup_auth(app: FastAPI) -> FastAPI:
    """
    Full authentication setup including all required middleware. Sets
    the authentication system based upon the `SETTINGS.auth_system`
    variable; if it's not "soauth" we default to the mock setup.
    """

    # Main authentication middleware. Note that this must come first
    # in the middleware stack so that it can set the user object
    # in the request scope.

    if SETTINGS.auth_system == "soauth":
        from soauth.toolkit.fastapi import global_setup

        from hipposerve.service.auth import AuthenticationError

        app = global_setup(
            app=app,
            app_base_url=SETTINGS.soauth_base_url,
            authentication_base_url=SETTINGS.soauth_service_url,
            app_id=SETTINGS.soauth_app_id,
            key_pair_type=SETTINGS.soauth_key_pair_type,
            public_key=SETTINGS.soauth_public_key,
            client_secret=SETTINGS.soauth_client_secret,
        )

        # Add exception handlers
        def redirect_to_login(request: Request, exc):
            if request.user.is_authenticated:
                return HTMLResponse(
                    "<p>You are not able to access this due to a lack of privileges</p>",
                )
            else:
                # Encode the URL to redirect to after login
                redirect_url = urlencode({"next": request.url.path})
                redirect_url = app.login_url + "?" + redirect_url
                return RedirectResponse(redirect_url)

        app.add_exception_handler(403, redirect_to_login)
        app.add_exception_handler(401, redirect_to_login)
        app.add_exception_handler(AuthenticationError, redirect_to_login)
    else:
        from soauth.toolkit.fastapi import mock_global_setup

        logger.warning(
            "Using mock authentication setup, this is not suitable for production use"
        )

        app = mock_global_setup(
            app=app, grants=["hippo:read", "hippo:write", "hippo:admin"]
        )

    # This middleware (above) is responsible for keeping the user
    # information up to date in the database.

    app.add_middleware(UserMiddleware)

    return app



from typing import Literal

from hippometa.base import BaseMetadata


class SimpleMetadata(BaseMetadata):
    """
    Simple metadata type for example purposes.
    """

    metadata_type: Literal["simple"] = "simple"



"""
A map set.
"""

from pathlib import Path
from typing import ClassVar, Literal

from hippometa.base import BaseMetadata


class MapSet(BaseMetadata):
    """
    A set of maps corresponding to the same observation. An easy way to package
    up e.g. a coadd mapp and its associated ivar map.
    """

    metadata_type: Literal["mapset"] = "mapset"
    valid_slugs: ClassVar[set[str]] = {
        "map",  # A regular, source and sky, and whatever else, map
        "source_only",  # A source-only map
        "source_free",  # A source-free map
        "ivar",  # Inverse-variance map
        "xlink",  # Cross-linking map
        "mask",  # Sky apodization mask
        "hits",  # Hit map (pixels that were hit by detectors)
        "weights",  # A weights map (?)
        "time",  # A time map, showing when individual detectors were hit
        "file_names",  # The file names of atomic maps or TODs used to create this map
        "data",  # Generic
    }

    pixelisation: Literal["equirectangular", "healpix", "cartesian"]

    telescope: str | None = None
    "Telescope that this map is created from"
    instrument: str | None = None
    "The instrument on that telescope that was used for this map"
    release: str | None = None
    "The data release this map is part of"
    season: str | None = None
    "The observing season(s) that are included in the map"
    patch: str | None = None
    "Sky patch observed in this map"
    frequency: str | None = None
    "On-sky frequency of the map (GHz), stored as a string to avoid round-off etc."
    polarization_convention: str | None = None
    "Polarization convention of the map (e.g. IAU)"

    split: str | int | None = None
    "The split that this map is of, or 'coadd' if it is a full map"

    tags: list[str] | None = None
    "Tags used to categorizet his map, including scan split strategies and so forth"

    @classmethod
    def from_fits(cls, filename: Path) -> "MapSet":
        """
        Load a MapSet from a FITS file.
        """
        from astropy.io import fits

        with fits.open(filename) as hdul:
            metadata = hdul[0].header
            pixelisation = metadata.get("PIXELIS", "healpix").lower()
            if pixelisation not in ["healpix", "equirectangular"]:
                # Check if CAR in ctype
                if "CTYPE1" in metadata and any("CAR" in metadata["CTYPE1"].upper()):
                    pixelisation = "equirectangular"
                else:
                    raise ValueError(f"Invalid pixelisation: {pixelisation}")

            return cls(
                filename=filename,
                pixelisation=pixelisation,
                telescope=metadata.get("TELESCOP"),
                instrument=metadata.get("INSTRUME"),
                release=metadata.get("RELEASE"),
                season=metadata.get("SEASON"),
                patch=metadata.get("PATCH"),
                frequency=metadata.get("FREQ", "").replace("f", ""),
                polarization_convention=metadata.get("POLCCONV", ""),
                tags=metadata.get("ACTTAGS", "").split(",")
                if metadata.get("ACTTAGS")
                else None,
            )
